#include <util/delay.h>
#include "analog_calibration.h"
#include "buttons.h"
#include "colors.h"
#include "crc.h"
#include "frame.h"
#include "hal_tx.h"


typedef struct configuration_t_
{
    analog_calibration_t analog_calibration[ANALOG_AXES_N];
}
configuration_t;


typedef enum mode_t_
{
    MODE_NORMAL,
    MODE_CALIBRATE_OUTPUT
}
mode_t;


static const FLASH configuration_t DEFAULT_CONFIGURATION =
{
    .analog_calibration = 
    {
        { .input_neutral = 512, .output_neutral = 512, .scale = 64 },
        { .input_neutral = 512, .output_neutral = 512, .scale = -64 },
        { .input_neutral = 512, .output_neutral = 512, .scale = 64 },
        { .input_neutral = 512, .output_neutral = 512, .scale = -64 },
    }
};


static configuration_t configuration;
static frame_t frame;
static uint32_t previous_buttons;
static uint32_t button_presses;
static rgb_t led_colors[LEDS_N];
static mode_t mode;
static analog_axis_t calibrated_axis;


static int16_t limit_range(int16_t value, int16_t min, int16_t max)
{
    if (value < min)
    {
        return min;
    }

    if (value > max)
    {
        return max;
    }

    return value;
}


static uint16_t read_analog(analog_axis_t axis)
{
    int16_t raw_sample = hal_tx_sample_axis(axis);
    int16_t output = analog_calibration_apply(&configuration.analog_calibration[axis], raw_sample);
    return limit_range(output, 0, 1023);
}


static uint32_t read_buttons(void)
{
    return hal_tx_read_buttons();
}


static void send_frame(frame_t const *frame)
{
    uint8_t const *frame_data = (uint8_t const *) frame;
    uint8_t frame_length = sizeof(frame_t);

    hal_tx_rf_send_start();

    for (uint8_t i = 0; i < frame_length; ++i)
    {
        for (uint8_t b = 0x80; b != 0; b >>= 1)
        {
            hal_tx_rf_send_bit(frame_data[i] & b);
        }
    }

    hal_tx_rf_send_stop();
}


static void leds_set(const rgb_t *color)
{
    for (uint8_t i = 0; i < LEDS_N; ++i)
    {
        led_colors[i] = *color;
    }
}


static void leds_flash(const rgb_t *color)
{
    leds_set(color);

    for (uint8_t i = 0; i < 40; ++i)
    {
        hal_tx_update_leds(led_colors, LEDS_N);
        _delay_ms(10);
    }
}


static void leds_blink(const rgb_t *color, uint8_t index)
{
    leds_set(&COLOR_BLACK);

    if ((hal_tx_get_time() & 0x100) == 0)
    {
        led_colors[index] = *color;
    }
}


static bool read_config()
{
    return hal_tx_read_from_nvm(&configuration, sizeof(configuration), 0);
}


static void write_config()
{
    hal_tx_write_to_nvm(&configuration, sizeof(configuration), 0);
}


static void reset_to_defaults()
{
    configuration = DEFAULT_CONFIGURATION;
}


static void mode_normal()
{
    leds_set(&COLOR_DARK_GREEN);

    const bool shape_pressed = (button_presses & (BUTTON_SQUARE | BUTTON_CIRCLE | BUTTON_TRIANGLE | BUTTON_X)) != 0;

    if ((frame.buttons & BUTTON_A) && shape_pressed)
    {
        mode = MODE_CALIBRATE_OUTPUT;

        switch (button_presses)
        {
            case BUTTON_SQUARE: // left
                calibrated_axis = ANALOG_AXIS_LEFT_H;
                break;

            case BUTTON_CIRCLE: // right
                calibrated_axis = ANALOG_AXIS_RIGHT_H;
                break;

            case BUTTON_TRIANGLE: // up
                calibrated_axis = ANALOG_AXIS_LEFT_V;
                break;

            case BUTTON_X: // down
                calibrated_axis = ANALOG_AXIS_RIGHT_V;
                break;
        }

        configuration.analog_calibration[calibrated_axis].input_neutral = hal_tx_sample_axis(calibrated_axis);
    }
    else if ((frame.buttons & (BUTTON_B | BUTTON_D)) && (button_presses & BUTTON_CIRCLE))
    {
        reset_to_defaults();
    }
}


static void cancel_calibration()
{
    mode = MODE_NORMAL;
    read_config();
    leds_flash(&COLOR_RED);
}


static void confirm_calibration()
{
    mode = MODE_NORMAL;
    write_config();
    leds_flash(&COLOR_GREEN);
}


static void mode_calibrate_output()
{
    analog_calibration_t *c = &configuration.analog_calibration[calibrated_axis];
    leds_blink(&COLOR_CYAN, calibrated_axis);

    switch (button_presses)
    {
        case BUTTON_UP:
            ++(c->output_neutral);
            break;

        case BUTTON_DOWN:
            --(c->output_neutral);
            break;

        case BUTTON_LEFT:
            ++(c->scale);
            break;

        case BUTTON_RIGHT:
            --(c->scale);
            break;

        case BUTTON_TRIANGLE:
            c->scale = -c->scale;
            break;

        case BUTTON_CIRCLE:
            cancel_calibration();
            break;
    
        case BUTTON_X:
            confirm_calibration();
            break;
    }
}


int main(void)
{
    hal_tx_init();

    if (read_config())
    {
        leds_flash(&COLOR_GREEN);
    }
    else
    {
        leds_flash(&COLOR_YELLOW);
        reset_to_defaults();
    }

    leds_set(&COLOR_BLACK);

    while (1)
    {
        hal_tx_update_leds(led_colors, LEDS_N);

        frame.analog1 = read_analog(ANALOG_AXIS_RIGHT_V);
        frame.analog2 = read_analog(ANALOG_AXIS_RIGHT_H);
        frame.analog3 = read_analog(ANALOG_AXIS_LEFT_V);
        frame.analog4 = read_analog(ANALOG_AXIS_LEFT_H);
        
        previous_buttons = frame.buttons;
        frame.buttons = read_buttons();
        button_presses = ~previous_buttons & frame.buttons;
        frame.toggles ^= button_presses;

        frame_update_crc(&frame);
        send_frame(&frame);

        switch (mode)
        {
            case MODE_NORMAL:
                mode_normal();
                break;

            case MODE_CALIBRATE_OUTPUT:
                mode_calibrate_output();
                break;
        }
    }
}