#include "hal_tx.h"
#include <stdbool.h>
#include <string.h>
#include <avr/eeprom.h>
#include <avr/io.h>
#include <util/delay.h>
#include "adc.h"
#include "colors.h"
#include "crc.h"


#define RF_OUTPUT_PORT PORTA
#define RF_OUTPUT_DDR DDRA
#define RF_OUTPUT_P PA5

#define LEDS_PORT PORTA
#define LEDS_DDR DDRA
#define LEDS_P PA0

#define BUTTONS_RESET_PORT PORTB
#define BUTTONS_RESET_DDR DDRB
#define BUTTONS_RESET_P PB0

#define BUTTONS_CLK_PORT PORTB
#define BUTTONS_CLK_DDR DDRB
#define BUTTONS_CLK_P PB1

#define BUTTONS_INPUT_A_PORT PORTA
#define BUTTONS_INPUT_A_PIN PINA
#define BUTTONS_INPUT_A_DDR DDRA
#define BUTTONS_INPUT_A_P PA7

#define BUTTONS_INPUT_B_PORT PORTB
#define BUTTONS_INPUT_B_PIN PINB
#define BUTTONS_INPUT_B_DDR DDRB
#define BUTTONS_INPUT_B_P PB2


static const analog_axis_t ANALOG_AXIS_TO_ADC_CHANNEL[ANALOG_AXES_N] = { 1, 2, 3, 4 };


void hal_tx_init()
{
    // Set clock to 8MHz if not set by fuse bits
    CLKPR = _BV(CLKPCE);
    CLKPR = 0;

    // Enable transmitter output
    RF_OUTPUT_DDR |= _BV(RF_OUTPUT_P);

    // Enable LEDs output
    LEDS_DDR |= _BV(LEDS_P);

    // Enable buttons outputs
    BUTTONS_RESET_DDR |= _BV(BUTTONS_RESET_P);
    BUTTONS_CLK_DDR |= _BV(BUTTONS_CLK_P);

    // Enable Timer 1 Normal Mode, f = 31250Hz
    TCCR1B = _BV(CS12);

    adc_init();
}


uint16_t hal_tx_get_time()
{
    return TCNT1 / 32;
}


uint16_t hal_tx_sample_axis(analog_axis_t axis)
{
    return adc_sample(ANALOG_AXIS_TO_ADC_CHANNEL[axis]);
}


void hal_tx_rf_send_start()
{
    RF_OUTPUT_PORT |= _BV(RF_OUTPUT_P);
    _delay_us(1000);
    RF_OUTPUT_PORT &= ~_BV(RF_OUTPUT_P);
    _delay_us(100);
}


void hal_tx_rf_send_stop()
{
    RF_OUTPUT_PORT |= _BV(RF_OUTPUT_P);
}


// original: 40/150, 100/70
void hal_tx_rf_send_bit(bool bit)
{
    RF_OUTPUT_PORT |= _BV(RF_OUTPUT_P);

    if (bit)
    {
        _delay_us(50);
        RF_OUTPUT_PORT &= ~_BV(RF_OUTPUT_P);
        _delay_us(160);
    }
    else
    {
        _delay_us(110);
        RF_OUTPUT_PORT &= ~_BV(RF_OUTPUT_P);
        _delay_us(80);
    }
}


void hal_tx_rf_transmit(uint8_t data)
{
    for (int i = 0x80; i != 0; i >>= 1)
    {
        if (i & data)
        {
            RF_OUTPUT_PORT &= ~_BV(RF_OUTPUT_P);
        }
        else
        {
            RF_OUTPUT_PORT |= _BV(RF_OUTPUT_P);
        }

        _delay_us(100);
    }
}


uint32_t hal_tx_read_buttons()
{
    uint32_t result = 0;

    BUTTONS_RESET_PORT |= _BV(BUTTONS_RESET_P);
    BUTTONS_RESET_PORT &= ~_BV(BUTTONS_RESET_P);

    for (uint8_t i = 0; i < 10; ++i)
    {
        const bool button_a = (BUTTONS_INPUT_A_PIN & _BV(BUTTONS_INPUT_A_P));
        const bool button_b = (BUTTONS_INPUT_B_PIN & _BV(BUTTONS_INPUT_B_P));

        BUTTONS_CLK_PORT |= _BV(BUTTONS_CLK_P);
        BUTTONS_CLK_PORT &= ~_BV(BUTTONS_CLK_P);

        result <<= 2;

        if (button_a)
        {
            result |= 1;
        }

        if (button_b)
        {
            result |= 2;
        }
    }

    return result;
}


#define LEDS_SEND_BIT(bit) \
        "sbi %[port], %[pin] \n\t" \
        "sbrs %[byte], " # bit " \n\t" \
        "cbi %[port], %[pin] \n\t" \
        "rjmp .+0 \n\t" \
        "rjmp .+0 \n\t" \
        "rjmp .+0 \n\t" \
        "sbrc %[byte], " # bit " \n\t" \
        "cbi %[port], %[pin] \n\t" \
        "nop \n\t"


void hal_tx_update_leds(rgb_t *data, uint8_t data_length)
{
    // 0: 3 high (0.375us), 11 low (1.375)
    // 1: 11 high (1.375), 3 low (0.375us)

    __asm__ volatile(
        
        "leds_write_start: \n\t"

        "ld __tmp_reg__, X+ \n\t"        // 2 0  2 1
        "sbrc %[byte], 0 \n\t"       // 2 0  1 1
        "cbi %[port], %[pin] \n\t"       // 0 0  2 0
        "mov %[byte], __tmp_reg__ \n\t"                       // 1 0  1 0

        LEDS_SEND_BIT(7)
        LEDS_SEND_BIT(6)
        LEDS_SEND_BIT(5)
        LEDS_SEND_BIT(4)
        LEDS_SEND_BIT(3)
        LEDS_SEND_BIT(2)
        LEDS_SEND_BIT(1)

        "sbi %[port], %[pin] \n\t"
        "sbrs %[byte], 0 \n\t"
        "cbi %[port], %[pin] \n\t"
        "subi %[remaining], 1 \n\t"
        "breq leds_write_finish \n\t"
        "rjmp leds_write_start \n\t"

        "leds_write_finish: \n\t"

        "rjmp .+0 \n\t" \
        "sbrc %[byte], 0" "\n\t"
        "cbi %[port], %[pin] \n\t"
        "nop \n\t"

        :
        : [port] "I" (_SFR_IO_ADDR(LEDS_PORT)),
          [pin] "I" (LEDS_P),
          [data] "x" (data),
          [remaining] "d" (data_length * sizeof(*data)),
          [byte] "d" (data[0])
        :
    );
}


bool hal_tx_read_from_nvm(void *target, uint8_t target_size, uint8_t index)
{
    uint16_t *crc = (uint16_t *) ((target_size + 2) * index);
    void *source = (void *) (crc + 1);
    eeprom_read_block(target, source, target_size);
    uint16_t actual_crc = crc16_ccitt_false_small(target, target_size);
    return (actual_crc == eeprom_read_word(crc));
}


void hal_tx_write_to_nvm(void *source, uint8_t source_size, uint8_t index)
{
    uint16_t *crc = (uint16_t *) ((source_size + 2) * index);
    void *target = (void *) (crc + 1);

    eeprom_update_word(crc, crc16_ccitt_false_small(source, source_size));
    eeprom_update_block(source, target, source_size);
}